﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=7.51
@EndOfDesignText@
'Custom View class
#Event: MouseMoved (EventData As MouseEvent)
'#DesignerProperty: Key: BooleanExample, DisplayName: Boolean Example, FieldType: Boolean, DefaultValue: True, Description: Example of a boolean property.
'#DesignerProperty: Key: IntExample, DisplayName: Int Example, FieldType: Int, DefaultValue: 10, MinRange: 0, MaxRange: 100, Description: Note that MinRange and MaxRange are optional.
'#DesignerProperty: Key: StringWithListExample, DisplayName: String With List, FieldType: String, DefaultValue: Sunday, List: Sunday|Monday|Tuesday|Wednesday|Thursday|Friday|Saturday
'#DesignerProperty: Key: StringExample, DisplayName: String Example, FieldType: String, DefaultValue: Text
'#DesignerProperty: Key: ColorExample, DisplayName: Color Example, FieldType: Color, DefaultValue: 0xFFCFDCDC, Description: You can use the built-in color picker to find the color values.
'#DesignerProperty: Key: DefaultColorExample, DisplayName: Default Color Example, FieldType: Color, DefaultValue: Null, Description: Setting the default value to Null means that a nullable field will be displayed.

Sub Class_Globals
	Private fx As JFX
	Private mEventName As String 'ignore
	Private mCallBack As Object 'ignore
	Private mBase As Pane
	
	'lib
	Private cv As Canvas
	Public FillHeight As Boolean = False
	Private VisibleRange(2) As Int
	Private PreviousDragPos(4) As Double 'posx, posy, relative ratio x, relative ratio y
	
	'Data
	Private Data As Map
	Private MinValue, MaxValue As Double
	Private Gradient As BitmapCreator
	Private GradientResolution As Int = 2049
	Public HighlightColor As Paint

End Sub

Public Sub Initialize (Callback As Object, EventName As String)
	mEventName = EventName
	mCallBack = Callback	
End Sub

Public Sub DesignerCreateView (Base As Pane, Lbl As Label, Props As Map)
	mBase = Base
	SetOpacity(mBase, 0)
	mBase.SetAlphaAnimated(2000, 1.0)
	
	Data.Initialize
	
	cv.Initialize("cv")
	cv.MouseCursor = fx.Cursors.NONE
	mBase.AddNode(cv, 0, 0, mBase.Width, mBase.Height)	
End Sub

Private Sub Base_Resize (Width As Double, Height As Double)
	cv.Width = Width
	cv.Height = Height
	
	UpdateChart(-1, -1)
End Sub

Public Sub GetBase As Pane
	Return mBase
End Sub

'The color gradient, from lower value to higher value.
'TIP: use xui.Color_ARGB to convert colors to int
Public Sub SetColorRange(Colors() As Int)	
	Gradient.Initialize(GradientResolution, 10)
	Dim r As B4XRect
	r.Initialize(0, 0, GradientResolution, 10)
	Gradient.FillGradient(Colors, r, "LEFT_RIGHT")
End Sub

Public Sub SetData(KeyValues As Map, ScaleValuesAccordingToMinMax As Boolean)
	Data = KeyValues
	VisibleRange = Array As Int(0, Data.Size -1)
	
	If ScaleValuesAccordingToMinMax Then
		Dim MinMax(2) As Double = GetMinMax
		MinValue = MinMax(0)
		MaxValue = MinMax(1)
	Else
		MinValue = -1.0
		MaxValue = 1.0
	End If	
	
	LogDebug($"Min Value: ${MinValue} | Max Value: ${MaxValue}"$)
End Sub

Public Sub GetMinValue As Double
	Return MinValue
End Sub

Public Sub GetMaxValue As Double
	Return MaxValue
End Sub

Public Sub GetScale As B4XBitmap
	Return Gradient.Bitmap
End Sub

Private Sub cv_MouseMoved (EventData As MouseEvent)
	UpdateChart(EventData.X, -1)
	PreviousDragPos(0) = 0
	PreviousDragPos(1) = 0
	PreviousDragPos(2) = -1
	PreviousDragPos(3) = -1
End Sub

Private Sub cv_MouseExited (EventData As MouseEvent)
	UpdateChart(-1, -1)
End Sub

Private Sub cv_MouseClicked (EventData As MouseEvent)
	If EventData.ClickCount = 2 And EventData.SecondaryButtonPressed Then
		VisibleRange = Array As Int(0, Data.Size -1)
		UpdateChart(-1, -1)
	End If
End Sub

Private Sub cv_MouseDragged (EventData As MouseEvent)
	If EventData.MiddleButtonPressed Then
		
		'record the zoom location
		If VisibleRange(1) - VisibleRange(0) <= Max(Data.Size * 0.01, 2) Then
			EventData.Consume
			Return
		End If
		
		If PreviousDragPos(2) = -1 Then
			PreviousDragPos(2) = EventData.X / cv.Width
		End If
		If PreviousDragPos(3) = -1 Then
			PreviousDragPos(3) = (cv.Width - EventData.X) / cv.Width
		End If
		
		Dim zoomAmount As Double = Max((VisibleRange(1) - VisibleRange(0)) * 0.01, 1)
		If EventData.X < PreviousDragPos(0) Then 'zoom in
			LogDebug($"Zoom in"$)
			VisibleRange(0) = Min(VisibleRange(0) + (zoomAmount * PreviousDragPos(2)), Data.Size -1)
			VisibleRange(1) = Max(VisibleRange(1) - (zoomAmount * PreviousDragPos(3)), 1)
		Else
			LogDebug($"Zoom out"$)
			VisibleRange(0) = Max(VisibleRange(0) - (zoomAmount * PreviousDragPos(2)), 1)
			VisibleRange(1) = Min(VisibleRange(1) + (zoomAmount * PreviousDragPos(3)), Data.Size -1)
		End If
		
		LogDebug($"Visible range: ${VisibleRange(0)} - ${VisibleRange(1)} zoom: ${zoomAmount} ratio: ${PreviousDragPos(2)}/${PreviousDragPos(3)}"$)
		
		UpdateChart(-1, -1)
		
		PreviousDragPos(0) = EventData.x
		PreviousDragPos(1) = EventData.y
		
	Else If EventData.SecondaryButtonPressed Then 'slide
		Dim SlideAmount As Int = PreviousDragPos(0) - EventData.X
		
		If SlideAmount > 0 Then 'slide right			
			If VisibleRange(1) + 1 > Data.Size -1 Then
				EventData.Consume
				Return
			End If	
			LogDebug($"slide right"$)
			VisibleRange(0) = VisibleRange(0) + 1
			VisibleRange(1) = VisibleRange(1) + 1
			
		Else
			
			If VisibleRange(0) - 1 < 0 Then
				EventData.Consume
				Return
			End If	
			LogDebug($"slide left"$)
			VisibleRange(0) = VisibleRange(0) - 1
			VisibleRange(1) = VisibleRange(1) - 1
			
		End If
		
		LogDebug($"Visible range: ${VisibleRange(0)} - ${VisibleRange(1)} (${VisibleRange(1) - VisibleRange(0)})"$)
		
		UpdateChart(-1, -1)
		
		PreviousDragPos(0) = EventData.x
		PreviousDragPos(1) = EventData.y
		
		
	End If	
End Sub

Private Sub cv_MouseReleased (EventData As MouseEvent)
	PreviousDragPos(0) = 0
	PreviousDragPos(1) = 0
	PreviousDragPos(2) = -1
	PreviousDragPos(3) = -1
End Sub

Private Sub SetOpacity(n As B4XView, Opacity As Double)
	Dim j As JavaObject = n
	j.RunMethod("setOpacity", Array(Opacity))
End Sub

Public Sub UpdateChart(HoverX As Double, FadeInDelayMs As Int)
	Dim Text As String
	
	If FadeInDelayMs > -1 Then
		SetOpacity(mBase, 0)
	End If
	
	cv.DrawRect(0, 0, cv.Width, cv.Height, fx.Colors.Black, True, 0)
	If Data.IsInitialized And Data.Size > 0 Then
		Dim LineWidth As Double = cv.Width / (VisibleRange(1) - VisibleRange(0))
		Dim Color As Paint
		Dim LineX As Double
		Dim x As Double
			
		For i = VisibleRange(0) To VisibleRange(1)
			x = ((i - VisibleRange(0)) * LineWidth)
			
			If HoverX > -1 And HoverX > x And HoverX < x + LineWidth Then
				Color = fx.Colors.Black
				LineX = x
				If HoverX < cv.Width / 2 Then
					Text = $"${Chr(0xF073)} ${Data.GetKeyAt(i)}${CRLF}${Chr(0xF2c8)} ${NumberFormat2(Data.GetValueAt(i), 1, 3, 3, False)} (C)"$
				Else	
					Text = $"${Data.GetKeyAt(i)} ${Chr(0xF073)}${CRLF}${NumberFormat2(Data.GetValueAt(i), 1, 3, 3, False)} (C) ${Chr(0xF2c8)}"$
				End If	
			Else
				If IsNaN(Data.GetValueAt(i)) Then
					Color = fx.Colors.Black
				Else
					Color = fx.Colors.From32Bit(Gradient.GetColor((GradientResolution -1) * Normalize(Data.GetValueAt(i)), 0))
				End If	
			End If
				
			If FillHeight Then
				cv.DrawRect(x, 0, LineWidth + 1, cv.Height, Color, True, 10)
			Else
				Dim LineHeight As Double = cv.Height * Normalize(Data.GetValueAt(i))
				cv.DrawRect(x, cv.Height - LineHeight, LineWidth + 1, LineHeight, Color, True, 10)
			End If	
		Next
		
		If Text <> "" Then
			Dim TextWidth As Double = 135
			Dim TextX As Double
			Dim Alignment As String
			Dim Font As Font = fx.CreateFontAwesome(15)
			Dim Height As Double = MeasureMultilineTextHeight(Font, TextWidth, Text)
			Dim Margin As Int = 10

			If HoverX > cv.Width / 2 Then 'right
				x = LineX - TextWidth - Margin * 2 + 1
				TextX = x + TextWidth
				Alignment = "RIGHT"
			Else 'left
				x = LineX + LineWidth - 1
				TextX = x
				Alignment = "LEFT"
			End If
			
			cv.DrawRect(x, (cv.Height / 2) - (Height / 2) - Margin, TextWidth + Margin* 2, Height + Margin *2 , fx.Colors.ARGB(128, 0, 0, 0), True, 0)
			cv.DrawText2(Text, TextX, cv.Height / 2, Font, fx.Colors.White, Alignment, TextWidth)
			
		End If
	End If
	
	If FadeInDelayMs > -1 Then
		mBase.SetAlphaAnimated(FadeInDelayMs, 1.0)		
	End If
	
End Sub

Private Sub IsNaN(Value As Double) As Boolean
	Dim j As JavaObject = Value
	Return j.RunMethod("isNaN", Null)
End Sub

Private Sub Normalize(Value As Double) As Double
	Return ((Value - MinValue) / (MaxValue - MinValue))
End Sub

Private Sub GetMinMax As Double(2)
	Dim dMin = 0, dMax = 0 As Double
	For Each d As Double In Data.Values
		If d < dMin Then
			dMin = d
		Else If d > dMax Then
			dMax = d	
		End If
	Next
	Return Array As Double(dMin, dMax)
End Sub

Sub MeasureMultilineTextHeight (Font As Font, Width As Double, Text As String) As Double
	Dim jo As JavaObject = Me
	Return jo.RunMethod("MeasureMultilineTextHeight", Array(Font, Text, Width))
End Sub
#if Java
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import javafx.scene.text.Font;
import javafx.scene.text.TextBoundsType;
public static double MeasureMultilineTextHeight(Font f, String text, double width) throws Exception {
  Method m = Class.forName("com.sun.javafx.scene.control.skin.Utils").getDeclaredMethod("computeTextHeight",
  Font.class, String.class, double.class, TextBoundsType.class);
  m.setAccessible(true);
  return (Double)m.invoke(null, f, text, width, TextBoundsType.LOGICAL_VERTICAL_CENTER);
  }
#End If